#include "renderer3D.hpp"

#include "vertex.hpp"
#include "vulkan_allocator.hpp"
#include "vulkan_instance.hpp"
#include "texture_manager.hpp"
#include "exceptions.hpp"
#include "vk_enum_string.h"
#include "vulkan_settings.hpp"

#include <cstring>
#include <chrono>

namespace gz::vlk {
//
// INIT & CLEANUP
//
    Renderer3D::Renderer3D(VulkanInstance& instance, TextureManager& textureManager) : 
        Renderer(instance, textureManager) {
        LogCreateInfo logCI{
        	.logfile = "renderer3D.log",
        	.storeLog = false,
        	.prefix = "3D-Renderer",
        	.prefixColor = Color::LI_CYAN,
        	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR,
        };
        rLog = Log(std::move(logCI));

        vk.registerCleanupCallback(std::bind(&Renderer3D::cleanup, this));
        vk.registerSwapChainRecreateCallback(std::bind(&Renderer3D::swapChainRecreateCallback, this));

        vk.createCommandBuffers(commandBuffers);
        const size_t vertexCount = 20000;
        const size_t indexCount = 10000;
        vk.createVertexBuffer<Vertex3D>(vertexCount, vertexBuffer, vertexBufferMemory, vertexBufferSize);
        vk.createIndexBuffer<uint32_t>(indexCount, indexBuffer, indexBufferMemory, indexBufferSize);

        // TODO
        loadModel();

        createUniformBuffers();
        createDescriptorResources();
        initSwapChainDependantResources();
        VulkanInstance::registerObjectUsingVulkan(ObjectUsingVulkan("Renderer3D",
                { &pipelines[PL_3D].pipeline, &renderPass, &vertexBuffer, &vertexBufferMemory, &indexBuffer, &indexBufferMemory,
                  &descriptorSetLayout, &descriptorPool, },
                { &framebuffers, &images, &imageMemory, &imageViews, &commandBuffers,
                  &descriptorSets }));
        rLog.log1("Created Renderer3D");
    }

    void Renderer3D::cleanup() {
        rLog.log0("cleanup: cleaning up");
        // UPDATE DOC ON CHANGES!
        for (size_t i = 0; i < vk.getMaxFramesInFlight(); i++) {
            vk.destroyBuffer(uniformBuffers[i], uniformBuffersMemory[i]);
        }
        vk.getDevice().destroyDescriptorSetLayout(descriptorSetLayout, NO_ALLOC);
        vk.getDevice().destroyDescriptorPool(descriptorPool, NO_ALLOC);

        cleanupSwapChainDependantResources();
        cleanup_();
    }


//
// SWAPCHAIN DEPENDANT
//
    void Renderer3D::initSwapChainDependantResources() {
        // UPDATE DOC ON CHANGES!
        createRenderPass();
        createImages();
        createDepthImage();
        vk.createFramebuffers(framebuffers, imageViews, renderPass, depthImageView);
        std::vector<vk::DescriptorSetLayout> descriptorSetLayouts = { descriptorSetLayout };
        vk.createGraphicsPipeline<Vertex3D>("shaders/vert.spv", "shaders/frag.spv", descriptorSetLayouts, true, renderPass, pipelines[PL_3D]);
    }


    void Renderer3D::cleanupSwapChainDependantResources() {
        // destroy pipelines
        pipelines.destroy(vk.getDevice());

        vk.destroyFramebuffers(framebuffers);

        vk.destroyImageView(depthImageView);
        vk.destroyImage(depthImage, depthImageMemory);

        for (size_t i = 0; i < images.size(); i++) {
            vk.destroyImageView(imageViews[i]);
            vk.destroyImage(images[i], imageMemory[i]);
        }
        vk.getDevice().destroyRenderPass(renderPass, nullptr);
    }


    void Renderer3D::swapChainRecreateCallback() {
        cleanupSwapChainDependantResources();
        initSwapChainDependantResources();
    }


//
// IMAGES
//
    void Renderer3D::createImages() {
        images.resize(vk.getScImages().size());
        imageMemory.resize(vk.getScImages().size());
        imageViews.resize(vk.getScImages().size());
        vk::ImageUsageFlags usage= vk::ImageUsageFlagBits::eColorAttachment | vk::ImageUsageFlagBits::eTransferSrc;
        for (size_t i = 0; i < images.size(); i++) {
            vk.createImage(vk.getScExtent().width, vk.getScExtent().height, vk.getScImageFormat(), vk::ImageTiling::eOptimal, usage, vk::MemoryPropertyFlagBits::eDeviceLocal, images[i], imageMemory[i]);
            vk.createImageView(vk.getScImageFormat(), images[i], imageViews[i], vk::ImageAspectFlagBits::eColor);
        }
    }


//
// DEPTH
//
    void Renderer3D::createDepthImage() {
        vk::MemoryPropertyFlags memoryProperties = vk::MemoryPropertyFlagBits::eDeviceLocal;
        vk::ImageUsageFlags imageUsage = vk::ImageUsageFlagBits::eDepthStencilAttachment;
        vk::ImageTiling tiling = vk::ImageTiling::eOptimal;
        vk::Format depthFormat = vk.getDepthFormat();

        vk.createImage(vk.getScExtent().width, vk.getScExtent().height, depthFormat, tiling, imageUsage, memoryProperties, depthImage, depthImageMemory);
        vk.createImageView(depthFormat, depthImage, depthImageView, vk::ImageAspectFlagBits::eDepth);

        vk.transitionImageLayout(depthImage, depthFormat, vk::ImageLayout::eUndefined, vk::ImageLayout::eDepthStencilAttachmentOptimal);
    }

//
// RENDER PASS
//
    void Renderer3D::createRenderPass() {
        vk::AttachmentDescription2 colorBlendAD {
            .format = vk.getScImageFormat(),
            .samples = vk::SampleCountFlagBits::e1,
            .loadOp = vk::AttachmentLoadOp::eClear,
            .storeOp = vk::AttachmentStoreOp::eStore,
            .stencilLoadOp = vk::AttachmentLoadOp::eDontCare,
            .stencilStoreOp = vk::AttachmentStoreOp::eDontCare,
            .initialLayout = vk::ImageLayout::eUndefined,
            .finalLayout = vk::ImageLayout::eTransferSrcOptimal,
        };

        vk::AttachmentReference2 colorBlendAR {
            .attachment = 0,
            .layout = vk::ImageLayout::eColorAttachmentOptimal,
        };

        vk::AttachmentDescription2 depthAD {
            .format = vk.getDepthFormat(),
            .samples = vk::SampleCountFlagBits::e1,
            .loadOp = vk::AttachmentLoadOp::eClear,
            .storeOp = vk::AttachmentStoreOp::eDontCare,
            .stencilLoadOp = vk::AttachmentLoadOp::eDontCare,
            .stencilStoreOp = vk::AttachmentStoreOp::eDontCare,
            .initialLayout = vk::ImageLayout::eUndefined,
            .finalLayout = vk::ImageLayout::eDepthStencilAttachmentOptimal,
        };

        vk::AttachmentReference2 depthAR {
            .attachment = 1,
            .layout = vk::ImageLayout::eDepthStencilAttachmentOptimal,
        };

        vk::SubpassDescription2 subpass {
            .pipelineBindPoint = vk::PipelineBindPoint::eGraphics,
            .colorAttachmentCount = 1,
            .pColorAttachments = &colorBlendAR,
            .pDepthStencilAttachment = &depthAR,
        };

        // wait until image is available
        vk::SubpassDependency2 colorAttachmentSD {
            .srcSubpass = VK_SUBPASS_EXTERNAL,
            .dstSubpass = 0,
            .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests,
            .dstStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests,
            .srcAccessMask = NO_ACC_FLAGS,
            .dstAccessMask = vk::AccessFlagBits::eColorAttachmentWrite | vk::AccessFlagBits::eDepthStencilAttachmentWrite,
        };

        // image layout transition to transfer dst
        vk::SubpassDependency2 layoutTransitionSD {
            .srcSubpass = 0,
            .dstSubpass = VK_SUBPASS_EXTERNAL,
            .srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput,
            .dstStageMask = vk::PipelineStageFlagBits::eTransfer,
            .srcAccessMask = vk::AccessFlagBits::eColorAttachmentWrite,
            .dstAccessMask = vk::AccessFlagBits::eTransferRead,
            .dependencyFlags = vk::DependencyFlagBits::eByRegion,
        };

        /* vk::SubpassDependency2 dependency { */
        /* dependency.srcSubpass = VK_SUBPASS_EXTERNAL; */
        /* dependency.dstSubpass = 0; */
        /* dependency.srcStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests; */
        /* dependency.srcAccessMask = 0; */
        /* dependency.dstStageMask = vk::PipelineStageFlagBits::eColorAttachmentOutput | vk::PipelineStageFlagBits::eEarlyFragmentTests; */ 
        /* dependency.dstAccessMask = vk::AccessFlagBits::eColorAttachmentWrite | vk::AccessFlagBits::eDepthStencilAttachmentWrite; */

        std::vector<vk::AttachmentDescription2> attachments = { colorBlendAD, depthAD, };
        std::vector<vk::SubpassDependency2> dependencies = { colorAttachmentSD, layoutTransitionSD };
        vk::RenderPassCreateInfo2 renderPassCI {
        	/* .attachmentCount = static_cast<uint32_t>(attachments.size()), */
        	/* .pAttachments = attachments.data(), */
        	/* .subpassCount = 1, */
        	/* .pSubpasses = &subpass, */
        	/* .dependencyCount = static_cast<uint32_t>(dependencies.size()), */
        	/* .pDependencies = dependencies.data(), */
            /* .dependencyCount = 0, */
            /* .pDependencies = nullptr, */
            /* .correlatedViewMaskCount = 0, */
            /* .pCorrelatedViewMasks = nullptr, */
        };
        renderPassCI.setAttachments(attachments);
        renderPassCI.setDependencies(dependencies);
        renderPassCI.setSubpasses(subpass);
        vk::Result result = vk.getDevice().createRenderPass2(&renderPassCI, nullptr, &renderPass); 
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Could not create render pass", "Renderer3D::createRenderPass");
        }
        rLog.log0("createRenderPass: Created render pass.");
    }



//
// DESCRIPTORS
//
    void Renderer3D::createDescriptorResources() {
        // LAYOUT
        // 1) uniform buffer object
        vk::DescriptorSetLayoutBinding uboLayoutBinding {
            .binding = 0,
            .descriptorType = vk::DescriptorType::eUniformBuffer,
            .descriptorCount = 1,
            .stageFlags = vk::ShaderStageFlagBits::eVertex,
            /* .pImmutableSamplers = nullptr, */
        };

        // 2) combined image sampler
        vk::DescriptorSetLayoutBinding samplerLayoutBinding {
            .binding = 1,
            .descriptorType = vk::DescriptorType::eCombinedImageSampler,
            .descriptorCount = 1,
            .stageFlags = vk::ShaderStageFlagBits::eFragment,
            /* .pImmutableSamplers = nullptr, */
        };

        std::vector<vk::DescriptorSetLayoutBinding> bindings = { uboLayoutBinding, samplerLayoutBinding };
        vk.createDescriptorSetLayout(bindings, descriptorSetLayout);

        // POOL
        std::array<vk::DescriptorPoolSize, 2> poolSizes;
        poolSizes[0].type = vk::DescriptorType::eUniformBuffer;
        poolSizes[0].descriptorCount = static_cast<uint32_t>(vk.getMaxFramesInFlight());
        poolSizes[1].type = vk::DescriptorType::eCombinedImageSampler;
        poolSizes[1].descriptorCount = static_cast<uint32_t>(vk.getMaxFramesInFlight());

        vk::DescriptorPoolCreateInfo poolCI {
        	.maxSets = static_cast<uint32_t>(vk.getMaxFramesInFlight()),
        };
        poolCI.setPoolSizes(poolSizes);

        vk::Result result;
        std::tie(result, descriptorPool) = vk.getDevice().createDescriptorPool(poolCI);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to create descriptor pool", "Renderer3D::createDescriptorResources");
        }

        // SETS
        std::vector<vk::DescriptorSetLayout> layouts(vk.getMaxFramesInFlight(), descriptorSetLayout);
        vk::DescriptorSetAllocateInfo setAI {
        	.descriptorPool = descriptorPool,
        };
        setAI.setSetLayouts(layouts);

        // TODO is resize needed?
        /* descriptorSets.resize(vk.getMaxFramesInFlight()); */
        std::tie(result, descriptorSets) = vk.getDevice().allocateDescriptorSets(setAI);
        if (result != vk::Result::eSuccess) {
           throw getVkException(result, "Failed to create descriptor sets", "Renderer3D::createDescriptorResources");
        }

        // configure sets
        for (size_t i = 0; i < vk.getMaxFramesInFlight(); i++) {
            vk::DescriptorBufferInfo bufferI {
            	.buffer = uniformBuffers[i],
            	.offset = 0,
            	.range = VK_WHOLE_SIZE,  // sizeof(UniformBufferObject),
            };

            vk::DescriptorImageInfo imageI {
            	.sampler = textureManager.getTextureAtlas().getTextureSampler(),
            	.imageView = textureManager.getTextureAtlas().getTextureImageView(),
            	.imageLayout = vk::ImageLayout::eShaderReadOnlyOptimal,
            };

            std::array<vk::WriteDescriptorSet, 2> descriptorW{};
            descriptorW[0].dstSet = descriptorSets[i];
            descriptorW[0].dstBinding = bindingUniformBuffer;
            descriptorW[0].dstArrayElement = 0;
            descriptorW[0].descriptorType = vk::DescriptorType::eUniformBuffer;
            descriptorW[0].setBufferInfo(bufferI);
            /* descriptorW[0].pImageInfo = nullptr; */
            /* descriptorW[0].pTexelBufferView = nullptr; */

            descriptorW[1].dstSet = descriptorSets[i];
            descriptorW[1].dstBinding = bindingCombinedImageSampler;
            descriptorW[1].dstArrayElement = 0;
            descriptorW[1].descriptorType = vk::DescriptorType::eCombinedImageSampler;
            descriptorW[1].setImageInfo(imageI);
            /* descriptorW[1].pTexelBufferView = nullptr; */
            
            // write 1, copy 0
            vk.getDevice().updateDescriptorSets(descriptorW, nullptr);
        }
        rLog.log0("createDescriptorResources: Created descriptor layouts, pool and sets.");
    }

//
// MODEL
//
	void Renderer3D::loadModel() {
        // load model into VerticesAndIndices struct
        rLog.log1("Renderer3D: loadModel: loading model");
        vk.loadModel(settings::MODEL_PATH, model);
        // TODO use correct type
        vk::DeviceSize requiredVertexBufferSize = model.vertices.size() * sizeof(Vertex3D);
        vk::DeviceSize requiredIndexBufferSize = model.indices.size() * sizeof(uint32_t);

        if (requiredVertexBufferSize > vertexBufferSize) { throw VkException("Renderer3D::loadModel: vertex buffer too small"); }
        if (requiredIndexBufferSize > indexBufferSize) { throw VkException("Renderer3D::loadModel: index buffer too small"); }

        rLog.log0("Renderer3D: loadModel: filling vertex buffer");
        // copy to vertexBuffer
        // create staging buffer
        vk::Buffer stagingBuffer;
        MemoryInfo stagingBufferMemory;
        vk.createBuffer(requiredVertexBufferSize, vk::BufferUsageFlagBits::eTransferSrc, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, stagingBuffer, stagingBufferMemory);
        // fill staging buffer
        void* data;
        vk::Result result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, requiredVertexBufferSize, NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer for vertices", "Renderer3D::loadModel");
        }

        memcpy(data, model.vertices.data(), requiredVertexBufferSize);
        vk.getDevice().unmapMemory(stagingBufferMemory.memory);
        // fill vertex buffer 
        vk.copyBuffer(stagingBuffer, vertexBuffer, requiredVertexBufferSize);
        vk.destroyBuffer(stagingBuffer, stagingBufferMemory);

        rLog.log0("Renderer3D: loadModel: filling index buffer");
        data = nullptr;
        // copy to index buffer
        vk.createBuffer(requiredIndexBufferSize, vk::BufferUsageFlagBits::eTransferSrc, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, stagingBuffer, stagingBufferMemory);
        // fill staging buffer
        result = vk.getDevice().mapMemory(stagingBufferMemory.memory, stagingBufferMemory.offset, requiredIndexBufferSize, NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer for indices", "Renderer3D::loadModel");
        }
        memcpy(data, model.indices.data(), requiredIndexBufferSize);
        vk.getDevice().unmapMemory(stagingBufferMemory.memory);
        // fill index buffer 
        vk.copyBuffer(stagingBuffer, indexBuffer, requiredIndexBufferSize);
        vk.destroyBuffer(stagingBuffer, stagingBufferMemory);
	}

//
// RENDERING
//
    void Renderer3D::recordCommandBuffer(uint32_t imageIndex, uint32_t currentFrame) {
        vk::CommandBufferBeginInfo commandBufferBI;
        vk::Result result = commandBuffers[currentFrame].begin(commandBufferBI); 
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to begin 3D command buffer", "Renderer3D::recordCommandBuffer");
        }

        // clear
        std::array<vk::ClearValue, 2> clearValues{};
        vk::ClearColorValue clearColor{};
        clearColor.setFloat32({0.1f, 0.0f, 0.0f, 0.5f});
        clearValues[0].setColor(std::move(clearColor));
        clearValues[1].setDepthStencil({1.0f, 0});

        vk::RenderPassBeginInfo renderPassBI {
        	.renderPass = renderPass,
        	.framebuffer = framebuffers[imageIndex],
            .renderArea { .offset = { 0, 0 }, .extent = vk.getScExtent() }
        };
        renderPassBI.setClearValues(clearValues);

        commandBuffers[currentFrame].beginRenderPass(renderPassBI, vk::SubpassContents::eInline);

        commandBuffers[currentFrame].bindPipeline(vk::PipelineBindPoint::eGraphics, pipelines[PL_3D].pipeline);
        vk::Buffer vertexBuffers[] = { vertexBuffer };
        vk::DeviceSize offsets[] = {0};
        uint32_t bindingCount = 1;
        uint32_t firstBinding = 0;
        commandBuffers[currentFrame].bindVertexBuffers(firstBinding, bindingCount, vertexBuffers, offsets);
        // TODO use correct index type!
        commandBuffers[currentFrame].bindIndexBuffer(indexBuffer, NO_OFFSET, vk::IndexType::eUint32);

        uint32_t firstSet = 0;
        commandBuffers[currentFrame].bindDescriptorSets(vk::PipelineBindPoint::eGraphics, pipelines[PL_3D].layout, firstSet, descriptorSets[currentFrame], {});

        int instanceCount = 1;
        int firstIndex = 0;
        int firstInstance = 0;
        commandBuffers[currentFrame].drawIndexed(static_cast<uint32_t>(model.indices.size()), instanceCount, firstIndex, NO_OFFSET, firstInstance);
        commandBuffers[currentFrame].endRenderPass();

        vk.copyImageToImage(commandBuffers[currentFrame], images[imageIndex], vk.getScImages()[imageIndex], vk.getScExtent());
        result = commandBuffers[currentFrame].end();
        if (result != vk::Result::eSuccess) {
            rLog.error("Failed to record 3D - command buffer", "VkResult:", result);
            throw getVkException(result, "Failed to record 3D - command buffer", "Renderer3D::recordCommandBuffer");
        }
        vk.submitThisFrame(commandBuffers[currentFrame]);
    }
//
// RENDERING
//
    void Renderer3D::drawFrame(uint32_t imageIndex) {
        commandBuffers[vk.getCurrentFrame()].reset(NO_CMD_RESET_FLAGS);
        /* recordCommandBuffer(imageIndex, vk.getCurrentFrame()); */
        updateUniformBuffer();
        recordCommandBuffer(imageIndex, vk.getCurrentFrame());
    }


    void Renderer3D::createUniformBuffers() {
        vk::DeviceSize bufferSize = sizeof(UniformBufferObject);

        uniformBuffers.resize(vk.getMaxFramesInFlight());
        uniformBuffersMemory.resize(vk.getMaxFramesInFlight());

        for (size_t i = 0; i < vk.getMaxFramesInFlight(); i++) {
            vk.createBuffer(bufferSize, vk::BufferUsageFlagBits::eUniformBuffer, vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent, uniformBuffers[i], uniformBuffersMemory[i]);
        }
    }


    void Renderer3D::updateUniformBuffer() {
        static auto startTime = std::chrono::high_resolution_clock::now();
        auto currentTime = std::chrono::high_resolution_clock::now();
        float time = std::chrono::duration<float, std::chrono::seconds::period>(currentTime - startTime).count();

        // TODO use push constant instead of ubo
        UniformBufferObject ubo{};
        ubo.model = glm::rotate(glm::mat4(1.0f), time * std::numbers::pi_v<float> / 2, glm::vec3(0.0f, 0.0f, 1.0f));
        ubo.view = glm::lookAt(glm::vec3(2.0f, 2.0f, 2.0f), glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f));
        ubo.projection = glm::perspective(glm::radians(45.0f), static_cast<float>(vk.getScExtent().width) / vk.getScExtent().height, 1.0f, 10.0f);
        /* ubo.model = glm::mat4(1); */
        /* ubo.view = glm::mat4(1); */
        /* ubo.projection = glm::mat4(1); */
        /* ubo.projection[1][1] *= -1;  // y coordinate inverted in opengl */
        MemoryInfo& uniformBufferMI = uniformBuffersMemory[vk.getCurrentFrame()];
        void* data;
        vk::Result result = vk.getDevice().mapMemory(uniformBufferMI.memory, uniformBufferMI.offset, sizeof(ubo), NO_MEM_FLAGS, &data);
        if (result != vk::Result::eSuccess) {
            throw getVkException(result, "Failed to map staging buffer", "Renderer3D::updateUniformBuffer");
        }
        memcpy(data, &ubo, sizeof(ubo));
        vk.getDevice().unmapMemory(uniformBufferMI.memory);
    }






}  // namespace gz::vlk
