#include "vulkan_util.hpp"
#include <iostream>
#include <vulkan/vulkan_core.h>
#include <vulkan/vulkan_enums.hpp>
#include <vulkan/vulkan_structs.hpp>
#define VULKAN_HPP_NO_CONSTRUCTORS
#include <vulkan/vulkan.hpp>


namespace gz::vlk {
// 
// CONVENIENCE
//
	PhysicalDeviceFeatures::PhysicalDeviceFeatures() {
		f.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2;
        f11.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_1_FEATURES;
        f12.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES;
        f13.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_3_FEATURES;
		f.pNext = &f11;
		f11.pNext = &f12;
		f12.pNext = &f13;
	}

	/* static_assert(sizeof(VkStructureType) != sizeof(vk::StructureType), "Objects do not have same size"); */
	/* static_assert(sizeof(VkPhysicalDeviceFeatures) != sizeof(vk::PhysicalDeviceFeatures), "Objects do not have same size"); */
	bool PhysicalDeviceFeatures::requiredFeaturesAvailable(const PhysicalDeviceFeatures& requiredFeatures) const {
		// iterate over all the features in the structs
		// VkPhysicalDeviceFeatures2
		size_t featureCount = sizeof(VkPhysicalDeviceFeatures) / sizeof(VkBool32);
		// pointer to first feature
		const VkBool32* pFeature = &f.features.robustBufferAccess;
		const VkBool32* pRequiredFeature = &requiredFeatures.f.features.robustBufferAccess;
		for (size_t i = 0; i < featureCount; i++) {
			if (*(pRequiredFeature + i) == VK_TRUE and *(pFeature + i) == VK_FALSE) {
				/* std::cout << "core feature not supported: i = " << i << "\n"; */
				return false;
			}
		}

		// VkPhysicalDeviceVulkan11Features
		// pointer to first feature: after sType and pNext
		featureCount = (sizeof(VkPhysicalDeviceVulkan11Features) - sizeof(VkStructureType) - sizeof(void*)) / sizeof(VkBool32);
		pFeature = &f11.storageBuffer16BitAccess;
		pRequiredFeature = &requiredFeatures.f11.storageBuffer16BitAccess;
		for (size_t i = 0; i < featureCount; i++) {
			if (*(pRequiredFeature + i) == VK_TRUE and *(pFeature + i) == VK_FALSE) {
				/* std::cout << "vulkan11 feature not supported: i = " << i << "\n"; */
				return false;
			}
		}

		// VkPhysicalDeviceVulkan12Features
		// pointer to first feature: after sType and pNext
		featureCount = (sizeof(VkPhysicalDeviceVulkan12Features) - sizeof(VkStructureType) - sizeof(void*)) / sizeof(VkBool32);
		pFeature = &f12.samplerMirrorClampToEdge;
		pRequiredFeature = &requiredFeatures.f12.samplerMirrorClampToEdge;
		for (size_t i = 0; i < featureCount; i++) {
			if (*(pRequiredFeature + i) == VK_TRUE and *(pFeature + i) == VK_FALSE) {
				/* std::cout << "vulkan12 feature not supported: i = " << i << "\n"; */
				return false;
			}
		}

		// VkPhysicalDeviceVulkan13Features
		// pointer to first feature: after sType and pNext
		featureCount = (sizeof(VkPhysicalDeviceVulkan13Features) - sizeof(VkStructureType) - sizeof(void*)) / sizeof(VkBool32);
		pFeature = &f13.robustImageAccess;
		pRequiredFeature = &requiredFeatures.f13.robustImageAccess;
		for (size_t i = 0; i < featureCount; i++) {
			if (*(pRequiredFeature + i) == VK_TRUE and *(pFeature + i) == VK_FALSE) {
				/* std::cout << "vulkan13 feature not supported: i = " << i << "\n"; */
				return false;
			}
		}
		return true;
	}

	void PipelineContainer::erase(const PipelineT& key, const vk::Device& device, const vk::AllocationCallbacks* pAllocator) {
        device.destroyPipeline(pipelines[key].pipeline, pAllocator);
        device.destroyPipelineLayout(pipelines[key].layout, pAllocator);
		pipelines.erase(pipelines.find(key));
	}
	PipelineContainer::iterator PipelineContainer::erase(const PipelineContainer::iterator& it, const vk::Device& device, const vk::AllocationCallbacks* pAllocator) {
        device.destroyPipeline(it->second.pipeline, pAllocator);
        device.destroyPipelineLayout(it->second.layout, pAllocator);
		return pipelines.erase(it);
		
	}
	void PipelineContainer::destroy(const vk::Device& device, const vk::AllocationCallbacks* pAllocator) {
        auto it = pipelines.begin(); while (it != pipelines.end()) {
            it = erase(it, device);
        }
	}


	vk::DependencyInfo getDepInfo(const vk::ImageMemoryBarrier2& barrier) {
		vk::DependencyInfo depI {
			.imageMemoryBarrierCount = 1,
			.pImageMemoryBarriers = &barrier,
		};
		return depI;
	}


	/* void getBufferMemoryRequirements(const vk::Device& device, const vk::Buffer& buffer, vk::BufferMemoryRequirementsInfo2& bufMemReq, vk::MemoryRequirements2& memReq) { */
	/* 	bufMemReq.pNext = nullptr; */
	/* 	bufMemReq.buffer = buffer; */

	/* 	memReq.pNext = nullptr; */
	/* 	vk::getBufferMemoryRequirements2(device, &bufMemReq, &memReq); */
	/* } */


	/* void getImageMemoryRequirements(const vk::Device& device, const vk::Image& image, vk::ImageMemoryRequirementsInfo2& imMemReq, vk::MemoryRequirements2& memReq) { */
	/* 	imMemReq.image = image; */

	/* 	memReq = device.getImageMemoryRequirements2(imMemReq); */
	/* } */


//
// 
//
	ObjectUsingVulkan::ObjectUsingVulkan(std::string&& name, std::vector<void*>&& ptrsToHandles, std::vector<void*>&& vectorWithPtrsToHandles) {
		objectName = std::move(name);
		for (auto it = vectorWithPtrsToHandles.begin(); it != vectorWithPtrsToHandles.end(); it++) {
			vectorWithPtrToHandle.emplace_back(reinterpret_cast<std::vector<void*>*>(*it));
		}
		ptrToHandle = std::move(ptrsToHandles);
		updateHandles();
	}

	void ObjectUsingVulkan::updateHandles() {
		handles.clear();
		for (auto it = ptrToHandle.begin(); it != ptrToHandle.end(); it++) {
			handles.insert(reinterpret_cast<uint64_t>(reinterpret_cast<void*>(*it)));
		}
		for (auto it = vectorWithPtrToHandle.begin(); it != vectorWithPtrToHandle.end(); it++) {
			for (auto it2 = (*it)->begin(); it2 != (*it)->end(); it2++) {
				handles.insert(reinterpret_cast<uint64_t>(reinterpret_cast<void*>(*it2)));
			}
		}
	}


} // namespace gz::vlk



